/*
 * Library interface to Linux x86 Performance-Monitoring Counters.
 *
 * Copyright (C) 1999  Mikael Pettersson
 */
#include <stdlib.h>
#include <errno.h>
#include <signal.h>
#include <unistd.h>
#include <sys/ptrace.h>
#include <sys/wait.h>
#include "asm/perfctr.h"
#define CONFIG_PERFCTR
#include "linux/prctl.h"
#include "libperfctr.h"

static int have_direct_access = 0;
static int init_done = 0;
static volatile struct perfctr_state kstate;	/* updated by kernel */

static int sys_perfctr(unsigned cmd, int whom, void *arg)
{
    if( !init_done ) {
	struct perfctr_info info;

	if( prctl(PR_PERFCTR, PERFCTR_INFO, -1, &info) < 0 )
	    return -1;
	/* Determine if we will be able to sample the counters
	 * ourselves, without the overhead of a system call.
	 * This will not be the case for pre-MMX Intel P5 (no
	 * RDPMC instruction) and IDT WinChip (PMCs and TSC
	 * are mutually exclusive) processors.
	 */
	have_direct_access = info.have_rdpmc && info.have_rdtsc;
	init_done = 1;
    }
    return prctl(PR_PERFCTR, cmd, whom, arg, 0);
}

int perfctr_info(struct perfctr_info *info)
{
    return sys_perfctr(PERFCTR_INFO, -1, info);
}

static __inline__ void rdtsc64(unsigned long long *u64)
{
    unsigned int *u32 = (unsigned int*)u64;
    __asm__ __volatile__("rdtsc " : "=a"(u32[0]), "=d"(u32[1]));
}

static __inline__ void rdpmc64(unsigned int ctr, unsigned long long *u64)
{
    unsigned int *u32 = (unsigned int*)u64;
    __asm__ __volatile__("rdpmc" : "=a"(u32[0]), "=d"(u32[1]) : "c"(ctr));
}

int perfctr_read_self(struct perfctr_state *state)
{
    if( have_direct_access && kstate.status > 0 ) {
	unsigned long long tsc_start, tsc, pmc[2];
	do {
	    tsc_start = kstate.counters.tsc_start;
	    rdtsc64(&tsc);
	    rdpmc64(0, &pmc[0]);
	    rdpmc64(1, &pmc[1]);
	    state->counters.tsc = kstate.counters.tsc + (tsc - tsc_start);
	    state->counters.pmc[0] = kstate.counters.pmc[0] + pmc[0];
	    state->counters.pmc[1] = kstate.counters.pmc[1] + pmc[1];
	} while( tsc_start != kstate.counters.tsc_start );
	state->counters.tsc_start = tsc_start;
	state->control = kstate.control;
	state->status = kstate.status;
	return 0;
    } else if( sys_perfctr(PERFCTR_READ, -1, state) < 0 ) {
	if( errno == ENXIO ) { /* no counters enabled */
	    memset(state, 0, sizeof *state);
	    return 0;
	} else
	    return -1;
    } else
	return 0;
}

int perfctr_control_self(struct perfctr_control *control)
{
    control->user = (struct perfctr_state*)&kstate;
    if( sys_perfctr(PERFCTR_CONTROL, -1, control) < 0 ) {
	kstate.status = PERFCTR_STATUS_STOPPED;
	return -1;
    }
    kstate.status = PERFCTR_STATUS_RUNNING;
    return 0;
}

int perfctr_stop_self(void)
{
    kstate.status = PERFCTR_STATUS_STOPPED;
    return sys_perfctr(PERFCTR_STOP, -1, NULL);
}

int perfctr_unlink_self(void)
{
    kstate.status = PERFCTR_STATUS_UNLINKED;
    return sys_perfctr(PERFCTR_UNLINK, -1, NULL);
}

int perfctr_attach(int pid)
{
    int fd;

    if( ptrace(PTRACE_ATTACH, pid, 0, 0) < 0 )
	return -1;
    if( waitpid(pid, NULL, WUNTRACED) < 0 )
	return -1;
    fd = sys_perfctr(PERFCTR_ATTACH, pid, NULL);
    ptrace(PTRACE_DETACH, pid, 0, SIGCONT);
    return fd;
}

int perfctr_read_attached(int fd, struct perfctr_state *state)
{
    return sys_perfctr(PERFCTR_READ, fd, state);
}

int perfctr_control_attached(int fd, struct perfctr_control *control)
{
    return sys_perfctr(PERFCTR_CONTROL, fd, control);
}

int perfctr_stop_attached(int fd)
{
    return sys_perfctr(PERFCTR_STOP, fd, NULL);
}

int perfctr_unlink_attached(int fd)
{
    return sys_perfctr(PERFCTR_UNLINK, fd, NULL);
}
